import asyncio
import json
import logging
import os
import queue
import sys
import threading
import time

from nats.aio.client import Client as NATS
from nats.aio.errors import NatsError

_app_name = ''
_appInfo = None
_product_sn = ''
_device_sn = ''

# get Config
_config_path = './etc/iotedge/config.json'
with open(_config_path, 'r') as load_f:
    try:
        load_dict = json.load(load_f)
        print(str(load_dict))
        # print('----config: {} -------'.format(load_dict))
        _app_name = load_dict['appName']
        _product_sn = load_dict['productSN']
        _device_sn = load_dict['deviceSN']

        if 'appInfo' in load_dict.keys():
            _appInfo = load_dict['appInfo']

    except Exception as e:
        print('load config file error:{}'.format(e))
        sys.exit(1)

print("application name:{}".format(_app_name))


class _Logger(object):
    def __init__(self, name):
        self.url = os.environ.get(
            'IOTEDGE_NATS_ADDRESS') or 'tcp://127.0.0.1:4222'
        self.nc = NATS()
        self.loop = asyncio.new_event_loop()
        self.queue = queue.Queue()
        self.name = name
        self.logger = logging.getLogger()
        format_str = logging.Formatter(
            '%(asctime)s - %(levelname)s: %(message)s')
        sh = logging.StreamHandler()
        sh.setFormatter(format_str)
        self.logger.addHandler(sh)

    async def _flush_cron(self):
        while True:
            await asyncio.sleep(20)
            await self.nc.flush()

    async def _conn(self):
        try:
            # if not self.nc.is_connected:
            await self.nc.connect(servers=[self.url], loop=self.loop)
            logging.debug('nats for logger connect success')
        except Exception as e1:
            logging.error(e1)
            sys.exit(1)

    async def _publish(self):
        while True:
            try:
                if not self.queue.empty():
                    msg = self.queue.get()
                    bty = json.dumps(msg)
                    await self.nc.publish(subject='edge.log.upload',
                                          payload=bty.encode('utf-8'))
                    await self.nc.flush()
                else:
                    await asyncio.sleep(0)
            except NatsError as e:
                logging.error(e)
            except Exception as e:
                logging.error(e)

    async def gather(self):
        await self._conn()
        tasks = [self._flush_cron(), self._publish()]
        await asyncio.gather(*tasks, loop=self.loop)

    def start(self):
        self.loop.run_until_complete(self.gather())

    def debug(self, msg):
        data = {
            'module': self.name,
            'level': 'debug',
            'message': msg,
            'timestamp': int(time.time()),
        }
        self.queue.put(data)
        self.logger.debug(msg)

    def info(self, msg):
        data = {
            'module': self.name,
            'level': 'info',
            'message': msg,
            'timestamp': int(time.time()),
        }
        self.queue.put(data)
        self.logger.info(msg)

    def error(self, msg):
        data = {
            'module': self.name,
            'level': 'error',
            'message': msg,
            'timestamp': int(time.time()),
        }
        self.queue.put(data)
        self.logger.error(msg)

    def warn(self, msg):
        data = {
            'module': self.name,
            'level': 'warn',
            'message': msg,
            'timestamp': int(time.time()),
        }
        self.queue.put(data)
        self.logger.warn(msg)

    def critical(self, msg):
        data = {
            'module': self.name,
            'level': 'critical',
            'message': msg,
            'timestamp': int(time.time()),
        }
        self.queue.put(data)
        self.logger.critical(msg)

    def setLevel(self, level):
        self.logger.setLevel(level)


_iotedge_logger = _Logger("app_"+_app_name)


def _init_logger():
    global _iotedge_logger
    _iotedge_logger.start()
    logging.debug('init logger success')


_t_logger = threading.Thread(target=_init_logger)
_t_logger.setDaemon(True)
_t_logger.start()


def getLogger():
    global _iotedge_logger
    return _iotedge_logger



import asyncio
import json
import os
import queue
import base64
import sys
import threading

from nats.aio.client import Client as NATS
from nats.aio.errors import NatsError
from iotedgeapplicationlinksdk import _app_name, getLogger, _product_sn, _device_sn, _appInfo

_logger = getLogger()


def exit_handler(signum, frame):
    sys.exit(0)


_msg_cb = None
_msg_rrpc_cb = None

_nat_publish_queue = queue.Queue()
_nat_client_publish_queue = queue.Queue()


class _natsPublish(object):
    def __init__(self):
        self.url = os.environ.get(
            'IOTEDGE_NATS_ADDRESS') or 'tcp://127.0.0.1:4222'
        self.nc = NATS()
        self.loop = asyncio.new_event_loop()

    async def _flush_cron(self):
        while True:
            await asyncio.sleep(20)
            await self.nc.flush()

    async def _conn(self):
        try:
            # if not self.nc.is_connected:
            await self.nc.connect(servers=[self.url], loop=self.loop)
        except Exception as e1:
            _logger.error(e1)
            sys.exit(1)

    async def _publish(self):
        global _nat_client_publish_queue

        while True:
            try:
                if not _nat_client_publish_queue.empty():
                    msg = _nat_client_publish_queue.get()
                    bty = msg['payload']
                    await self.nc.publish(subject=msg['subject'],
                                          payload=bty.encode('utf-8'))
                    await self.nc.flush()
                else:
                    await asyncio.sleep(0)

            except NatsError as e:
                _logger.error(e)
            except Exception as e:
                _logger.error(e)

    async def gather(self):
        await self._conn()
        tasks = [self._flush_cron(), self._publish()]
        await asyncio.gather(*tasks, loop=self.loop)

    def start(self):
        self.loop.run_until_complete(self.gather())


class _natsClientPub(object):
    def __init__(self):
        self.url = os.environ.get(
            'IOTEDGE_NATS_ADDRESS') or 'tcp://127.0.0.1:4222'
        self.nc = NATS()
        self.loop = asyncio.new_event_loop()

    async def _flush_cron(self):
        while True:
            await asyncio.sleep(20)
            await self.nc.flush()

    async def _conn(self):
        try:
            # if not self.nc.is_connected:
            await self.nc.connect(servers=[self.url], loop=self.loop)
        except Exception as e1:
            _logger.error(e1)
            sys.exit(1)

    async def _publish(self):
        global _nat_publish_queue

        while True:
            try:
                if not _nat_publish_queue.empty():
                    msg = _nat_publish_queue.get()
                    bty = json.dumps(msg['payload'])
                    await self.nc.publish(subject=msg['subject'],
                                          payload=bty.encode('utf-8'))
                    await self.nc.flush()
                else:
                    await asyncio.sleep(0)

            except NatsError as e:
                _logger.error(e)
            except Exception as e:
                _logger.error(e)

    async def gather(self):
        await self._conn()
        tasks = [self._flush_cron(), self._publish()]
        await asyncio.gather(*tasks, loop=self.loop)

    def start(self):
        self.loop.run_until_complete(self.gather())


class _natsSubscribe(object):
    def __init__(self, subject: str, queue: str, cb):
        self.url = os.environ.get(
            'IOTEDGE_NATS_ADDRESS') or 'tcp://127.0.0.1:4222'
        self.nc = NATS()
        self.loop = asyncio.new_event_loop()
        self.cb = cb
        self.subject = subject
        self.queue = queue

    async def _flush_cron(self):
        while True:
            await asyncio.sleep(20)
            await self.nc.flush()

    async def _conn(self):
        try:
            # if not self.nc.is_connected:
            await self.nc.connect(servers=[self.url], loop=self.loop)
        except Exception as e1:
            _logger.error(e1)
            sys.exit(1)

    async def _sub(self):
        await self.nc.subscribe(self.subject, queue=self.queue, cb=self.cb, is_async=True)
        await self.nc.flush()

    async def gather(self):
        await self._conn()
        tasks = [self._flush_cron(), self._sub()]
        await asyncio.gather(*tasks, loop=self.loop)

    def start(self):
        self.loop.run_until_complete(self.gather())
        self.loop.run_forever()


class _natsClientSub(object):
    def __init__(self):
        self.url = os.environ.get(
            'IOTEDGE_NATS_ADDRESS') or 'tcp://127.0.0.1:4222'
        self.nc = NATS()
        self.loop = asyncio.new_event_loop()

    async def _flush_cron(self):
        while True:
            await asyncio.sleep(20)
            await self.nc.flush()

    async def _conn(self):
        try:
            # if not self.nc.is_connected:
            await self.nc.connect(servers=[self.url], loop=self.loop)
        except Exception as e1:
            _logger.error(e1)
            sys.exit(1)

    async def _sub(self):
        async def message_handler(msg):
            global _msg_rrpc_cb
            global _msg_cb
            _logger.debug("recv message:{} ".format(str(msg)))
            try:
                js = json.loads(msg.data.decode())
                topic = js['topic']
                data = base64.b64decode(js['payload'])
                if isinstance(topic, str) and topic.startswith("/$system/") and topic.count('/rrpc/request/') > 0:
                    topic = topic.replace("/request/", "/response/", 1)
                    if _msg_rrpc_cb:
                        _msg_rrpc_cb(topic, data)
                        return

                if _msg_cb:
                    _msg_cb(topic, data)
            except Exception as e:
                _logger.error('handle msg error {}'.format(str(e)))

        await self.nc.subscribe("edge.app." + _app_name, queue=_app_name, cb=message_handler, is_async=True)
        await self.nc.flush()

    async def gather(self):
        await self._conn()
        tasks = [self._flush_cron(), self._sub()]
        await asyncio.gather(*tasks, loop=self.loop)

    def start(self):
        self.loop.run_until_complete(self.gather())
        self.loop.run_forever()


def register_callback(cb, rrpc_cb):
    global _msg_cb
    _msg_cb = cb

    global _msg_rrpc_cb
    _msg_rrpc_cb = rrpc_cb


def get_gateway_product_sn():
    return _product_sn


def get_gateway_device_sn():
    return _device_sn


def get_application_name():
    return _app_name


def get_application_config():
    return _appInfo


def publish(topic: str, msg: bytes):
    global _nat_publish_queue
    payload_encode = base64.b64encode(msg)
    payload = {
        'src': "app",
        'topic': topic,
        "identity": _app_name,
        'payload': str(payload_encode, encoding='utf-8')
    }
    data = {
        'subject': 'edge.router.' + _app_name,
        'payload': payload
    }
    _nat_publish_queue.put(data)


def natsPublish(subject: str, payload: bytes):
    global _nat_client_publish_queue
    data = {
        'subject': subject,
        'payload': payload
    }
    _nat_client_publish_queue.put(data)


def natsSubscribe(subject, queue, cb):
    def _nats_sub():
        _natsSubscribe(subject, queue, cb).start()

    t = threading.Thread(target=_nats_sub)
    t.setDaemon(True)
    t.start()


def _start_pub():
    _natsClientPub().start()


def _nats_pub():
    _natsPublish().start()


def _start_sub():
    _natsClientSub().start()


_t_pub = threading.Thread(target=_nats_pub)
_t_pub.setDaemon(True)
_t_pub.start()

_t_nats_pub = threading.Thread(target=_start_pub)
_t_nats_pub.setDaemon(True)
_t_nats_pub.start()

_t_nats_sub = threading.Thread(target=_start_sub)
_t_nats_sub.setDaemon(True)
_t_nats_sub.start()



import sys

from setuptools import setup

if not (sys.version_info[0] == 3):
    sys.exit("Link IoT Edge only support Python 3")

setup(
    name='iotedge_application_link_sdk',
    version='0.3.0',
    author='ucloud.cn',
    url='https://pypi.org/project/iotedge_application_link_sdk/',
    author_email='ibu-bot@ucloud.cn',
    packages=['iotedgeapplicationlinksdk'],
    platforms="any",
    license='Apache 2 License',
    install_requires=[
        "asyncio-nats-client>=0.10.0",
        "cachetools>=4.0.0"
    ],
    description="IoT Stack Edge application Link SDK",
    long_description="IoT Stack Edge application Link SDK\n https://docs.ucloud.cn/uiot-stack/README"
)
